with Ada.Text_IO;              use Ada.Text_IO;
with Ada.Command_Line;         use Ada.Command_Line;
with Wordcount_Maps;          
with Wordcount_Lists;          
with Ada.Characters.Handling;  use Ada.Characters.Handling;   
with Ada.Integer_Text_IO;      use Ada.Integer_Text_IO;
with Ada.Strings.Unbounded;    use Ada.Strings.Unbounded;
with Ada.Strings.Maps;         use Ada.Strings.Maps;
with Ada.Strings.Fixed;        use Ada.Strings.Fixed;
with Charles.Unbounded_Lists.Generic_Quicksort;

procedure Main is

   Map : Wordcount_Maps.Map_Type;

   type Natural_Access is access all Natural;
   for Natural_Access'Storage_Size use 0;
   
   function Get_Count is 
      new Wordcount_Maps.Generic_Item (Natural_Access);

   procedure Increment (Word : String) is   
      use Wordcount_Maps;

      Iterator : Iterator_Type;
      Inserted : Boolean;
   begin
      Insert (Map, To_Lower (Word), Iterator, Inserted);

      declare
         Count : Integer renames Get_Count (Iterator).all;
      begin
         if Inserted then
            Count := 0;
         end if;
         
         Count := Count + 1;
      end;
   end Increment;      
      
   File : File_Type;
   
   Word_Chars : constant Character_Set :=
     To_Set (Character_Ranges'(('a', 'z'), ('A', 'Z')));
   
   Line : String (1 .. 2000);
  
   Line_First : Positive;
   Line_Last : Natural;

   Word_First : Positive;
   Word_Last : Natural;

   List : Wordcount_Lists.List_Type;

begin

   if Argument_Count = 0 then
      Put_Line (Command_Name & " <file>");
      return;
   end if;
   
   if Argument_Count > 1 then
      Put_Line ("too many command-line arguments");
      return;
   end if;
   
   begin
      Open (File, In_File, Name => Argument (1));
   exception
      when Name_Error =>
         Put_Line ("unable to open file");
         return;
   end;
   
   while not End_Of_File (File) loop

      Get_Line (File, Line, Line_Last);
      pragma Assert (Line_Last < Line'Last);

      Line_First := Line'First;

      loop

         Find_Token 
           (Source => Line (Line_First .. Line_Last),
            Set    => Word_Chars,
            Test   => Ada.Strings.Inside,
            First  => Word_First,
            Last   => Word_Last);

         exit when Word_Last = 0;
         
         Increment (Line (Word_First .. Word_Last));
         
         Line_First := Word_Last + 1;

      end loop;
            
   end loop;
   
   declare
      use Wordcount_Maps;

      I : Iterator_Type := First (Map);
      B : constant Iterator_Type := Back (Map);
      
      use Wordcount_Lists;
   begin
      while I /= B loop
         Push_Back (List, I);
         I := Succ (I);
      end loop;
   end;
   
   declare
      use Wordcount_Maps;

      function "<" (L, R : Iterator_Type) 
         return Boolean is
         
         LI : constant Natural := Item (L);
         RI : constant Natural := Item (R);
      begin
         return LI > RI;  --yes
      end;
      
      procedure Sort is new Wordcount_Lists.Generic_Quicksort;
   begin
      Sort (List);
   end;
  
--   declare
--      use Wordcount_Lists;

--      I : Iterator_Type := First (List);
--      B : constant Iterator_Type := Back (List);
--   begin
--      while I /= B loop
--         Put (To_String (Key (I)));
--         Put (' ');
--         Put (Item (I), Width => 0);
--         New_Line;
--         
--         I := Succ (I);
--      end loop;
--   end;

   declare
      use Wordcount_Lists;

      LI : Iterator_Type := First (List);
      B : constant Iterator_Type := Back (List);
      
      MI : Wordcount_Maps.Iterator_Type;
      
      use Wordcount_Maps;
   begin
      for J in 1 .. 10 loop
         exit when LI = B;
         
         MI := Item (LI);

         Put (To_String (Key (MI)));
         Put (' ');
         Put (Item (MI), Width => 0);
         New_Line;
         
         LI := Succ (LI);
      end loop;
   end;

 
end Main;

